import sys
import time
import evdev
import uinput
import logging
import traceback
from waylandmap.filter import Filter
from waylandmap.constants import KEYS_VALUE_TUPLE, code_to_name
from waylandmap.devices import get_device_path


def log_input_output(input, output=None):
    timestamp_in, type_in, code_in, value_in = input
    name_in = code_to_name(code_in) if type_in == 1 else ''
    if output is not None:
        (type_out, code_out), value_out = output
        name_out = code_to_name(code_out) if type_out == 1 else ''
    else:
        # if discarded by filter
        name_out = type_out = code_out = value_out = ''
    # log
    if type_in == 0:
        logging.info(f'\rtime: {timestamp_in}\t---- SYNC_REPORT ----')
    else:
        logging.info((
            f'\rtime: {timestamp_in}'
            f'\tIN: ({name_in:>15},T={type_in:1},C={code_in:3},V={value_in:6}), '
            f'\tOUT: ({name_out:>15},T={type_out:1},C={code_out:3},V={value_out:6})'))


def infinite_retry(sleep, catch):
    def wrapper(func):
        def wrapped(*args, **kwargs):
            while True:
                try:
                    func(*args, **kwargs)
                except PermissionError:
                    logging.error("Must be run as sudo")
                except catch as e:
                    logging.error(f'{str(e)}: Failed to connect to device, possibly due to wake from sleep, will keep retrying ...')
                    logging.error(traceback.format_exc())
                    time.sleep(sleep)
        return wrapped
    return wrapper


@infinite_retry(sleep=1, 
                catch=(FileNotFoundError, OSError, Exception))
def run(dev_name, keymaps):
    filter = Filter(keymaps)
    # start capturing from evdev
    kb = evdev.InputDevice(get_device_path(dev_name))
    kb.grab()

    try:
        with uinput.Device(KEYS_VALUE_TUPLE) as vdev:
            # loop to capture every event
            for ev in kb.read_loop():
                # each event represented by three ints
                _, type_in, code_in, value_in = input = ev.timestamp(), ev.type, ev.code, ev.value
                # ask filter to for correct mapping
                output = filter.target(type_in, code_in, value_in)
                # this key is a registered modifier key
                if output is None:
                    # drop the event and move on to next
                    log_input_output(input)
                    continue
                # emit the event to OS
                vdev.emit(*output)
                # log
                log_input_output(input, output)

    except KeyboardInterrupt:
        kb.ungrab()
        logging.error('User keyboard interrupted, quitting now.')
        sys.exit(0)




from evdev.ecodes import EV_KEY
import yaml
from waylandmap.constants import name_to_code as ntc


class Filter:
    def __init__(self, path: str):
        with open(path) as f:
            keymaps = yaml.safe_load(f)
        # organize into swap list and map list
        self._map = tuple((ntc(m['source']), ntc(m['target']))
                          for m in keymaps if m['type'] == 'map')
        self._swap = tuple((ntc(m['target1']), ntc(m['target2']))
                           for m in keymaps if m['type'] == 'swap')
        self._combo = tuple((ntc(m['modifier']), ntc(m['source']), ntc(m['target']))
                            for m in keymaps if m['type'] == 'combo')
        # modifier keys states
        self._on_modifier = {k[0]:False for k in self._combo}
        # mapping path states
        self._on_combo = {c:False for c in self._combo}

    def target(self, type_in: int, code_in: int, value_in: int) -> tuple | None:
        type_out, code_out, value_out = type_in, code_in, value_in
        # only interested in key event, ignore sync event
        if type_in == EV_KEY:
            # handle map
            for map in self._map:
                source, target = map
                if code_in == source:
                    code_out = target
                    return ((type_out, code_out), value_out)
            # handle swap
            for swap in self._swap:
                target1, target2 = swap
                if code_in == target1:
                    code_out = target2
                    return ((type_out, code_out), value_out)
                if code_in == target2:
                    code_out = target1
                    return ((type_out, code_out), value_out)

            ## handle combo
            # check if input is a modifier key
            if code_in in self._on_modifier:
                # if confirm, change on_mods state
                self._on_modifier[code_in] = value_in >= 1
                # then issue discard signal
                return None
            # check all registered mods tree and trigger
            for remap_path in self._combo:
                modifier, source, target = remap_path
                # check if modifier is on, or on_mapping flag is on
                if self._on_modifier[modifier] or self._on_combo[remap_path]:
                    # check if key pressed matched source
                    if code_in == source:
                        # change output target accordingly
                        code_out = target
                        # mark remap path flag to on, only turn off upon key release event
                        self._on_combo[remap_path] = value_in >= 1
                        # only need to match the first path
                        return ((type_out, code_out), value_out)
        # unless is a registered modifier key, always return a valid result
        return ((type_out, code_out), value_out)




import uinput

# {'ITEM_XXX': (0,0)}
EVENTS_NAME_VALUE_DICT = {
    k:v for k,v in vars(uinput.ev).items()
    if k.startswith(('KEY_', 'BTN_', 'REL_', 'ABS_', ))}

# {'KEY_XXX': (0,0)}
KEYS_NAME_VALUE_DICT = {
    k:v for k,v in EVENTS_NAME_VALUE_DICT.items()
    if k.startswith('KEY_')}

# {(0,0), ...}
KEYS_VALUE_TUPLE = KEYS_NAME_VALUE_DICT.values()

# {(0,0): 'KEY_XXX'}
KEYS_VALUE_NAME_DICT = {
    v:k for k,v in KEYS_NAME_VALUE_DICT.items()}

# {'KEY_XXX': 0}
KEY_NAME_CODE_DICT = {
    k:v[1] for k,v in KEYS_NAME_VALUE_DICT.items()}

# {0: 'KEY_XXX'}
KEY_CODE_NAME_DICT = {
    v:k for k,v in KEY_NAME_CODE_DICT.items()}

def name_to_code(name: str) -> int:
    code = KEY_NAME_CODE_DICT[name]
    return code

def code_to_name(code: int) -> str:
    name = KEY_CODE_NAME_DICT[code]
    return name



import evdev


def get_devices() -> list:
    devices = [evdev.InputDevice(dev) for dev in evdev.list_devices()]
    return devices

def get_device_path(name):
    for dev in get_devices():
        if name == dev.name:
            return dev.path
    else:
        raise FileNotFoundError('Failed to look up device name.')




import click
import logging
from waylandmap.devices import get_devices
from waylandmap.keymapper import run


@click.command(short_help='WaylandMap is a keymapper tools that works on both X11 or Wayland machine.')
@click.option('-l', '--list-devices', is_flag=True, default=False, help='List name of all available devices. Please find you keyboard.')
@click.option('-n', '--name', default=None, help='Name of the target keyboard.')
@click.argument('keymaps', nargs=1, required=False, default=None)
@click.option('-v', '--verbose', is_flag=True, default=False, help='Print live event mapping info to terminal.')
def cli(list_devices, name, keymaps, verbose):
    """KEYMAPS is the path to your config yaml file"""
    # show available devices
    if list_devices:
        devs = get_devices()
        if len(devs) == 0:
            click.echo('Need root permission to access devices.')
        else:
            print('List of avaliable devices:')
            for dev in devs:
                print(f'{dev.path}\t\t{dev.name}')
        return
    # ensure device name
    if name is None:
        print('Please provide the name of you keyboard to remap. You may use the `-l` flag to show all available devices.')
        return
    else:
        devs = get_devices()
        for dev in devs:
            if name == dev.name:
                dev_name = dev.name
                break
        else:
            print('Name of devices does not exists. Please check your device name.')
            return
    # ensure keymap config file
    if keymaps is None:
        print('Please provide the path to your keymap config file.')
        return
    # set log level
    logging.basicConfig(format='%(message)s',
                        level=logging.DEBUG if verbose else logging.WARNING)
    # finally run the program
    run(dev_name, keymaps)

